/****************************************************************************/
/* VICScript                                                                */
/* Simple programming language interpreter.                                 */
/*                                                                          */
/* Uses shunting yard algorithm for parsing expressions with operator       */
/* precedence:                                                              */
/* Ref: http://www.engr.mun.ca/~theo/Misc/exp_parsing.htm#shunting_yard     */
/*                                                                          */
/* Compiles with cc65 v2.15 for VIC-20 (and gcc 6.2 for Linux)              */
/*                                                                          */
/* cc65: Define symbol C64 to build for Commodore 64. Otherwise VIC-20 is   */
/*       assumed.                                                           */
/*                                                                          */
/* Copyright Bobbi Webber-Manners 2016, 2017                                */
/****************************************************************************/

/****************************************************************************/
/*                                                                          */
/*    This program is free software: you can redistribute it and/or modify  */
/*    it under the terms of the GNU General Public License as published by  */
/*    the Free Software Foundation, either version 3 of the License, or     */
/*    (at your option) any later version.                                   */
/*                                                                          */
/*    This program is distributed in the hope that it will be useful,       */
/*    but WITHOUT ANY WARRANTY; without even the implied warranty of        */
/*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         */
/*    GNU General Public License for more details.                          */
/*                                                                          */
/*    You should have received a copy of the GNU General Public License     */
/*    along with this program.  If not, see <http://www.gnu.org/licenses/>. */
/*                                                                          */
/****************************************************************************/
 
//#include <stdio.h>
#include <ctype.h>
#include <stdlib.h>
#include <string.h>
#include <setjmp.h>
#include <unistd.h>

#ifndef __GNUC__
#include <cbm.h>
#include <peekpoke.h>
#endif

//#define TEST
//#define DEBUG_READFILE

//#define EXIT(arg) {printf("%d\n",__LINE__); exit(arg);}
#define EXIT(arg) exit(arg)

#ifndef __GNUC__
#define leftarrow 0x5f /* CBM left arrow character */
#else
#define leftarrow '_'  /* ASCII underscore */
#endif

/*
 ***************************************************************************
 * Lightweight functions (lighter than the cc65 library for our use case)
 ***************************************************************************
 */

/*
 * This does the same thing as fputs(str, 1), but uses marginally less
 * memory.
 */
void print(char *str) {
	write(1, str, strlen(str));
}

/*
 * This does the same thing as printchar() but uses marginally less memory.
 */
void printchar(char c) {
	write(1, &c, 1);
}

/*
 * This should use less memory than the table-driven version of isalpha()
 * provided by cc65
 */
unsigned char isalphach(char ch) {
	return (ch >= 'a' && ch <= 'z') || (ch >= 'A' && ch <= 'Z');
}

/*
 * This should use less memory than the table-driven version of isdigit()
 * provided by cc65
 */
unsigned char isdigitch(char ch) {
	return (ch >= '0' && ch <= '9');
}

/*
 * Implement this ourselves for cc65 - x^y
 */
int _pow(int x, int y) {
	int i;
	int ret = 1;

	for (i = 0; i < y; i++) {
		ret *= x;
	}
	return ret;
}

/*
 * This is lighter than gets() and also safe!
 * Will read up to buflen bytes from STDIN
 */
void getln(char *str, unsigned char buflen) {
	unsigned char i;
	unsigned char j = 0;
	do {
		i = read(0, str + j++, 1);
	} while ((i != 0) && (j < buflen) && *(str + j - 1) != '\n');
	str[j - 1] = '\0';
}

/*
 ***************************************************************************
 * Prototypes
 ***************************************************************************
 */
unsigned char P(void);
unsigned char E(void);
unsigned char eval(unsigned char checkNoMore, int *val);
unsigned char parseint(int*);
unsigned char parsehexint(int*);
void printdec(unsigned int val);
void printhex(unsigned int val);
unsigned char getintvar(char*, int, int*);
unsigned char openfile(unsigned char);
unsigned char readfile(void);
unsigned char writefile(void);
void list(unsigned int, unsigned int);
void run(void);
void new(void);
unsigned char parseline(void);

/*
 ***************************************************************************
 * Globals
 ***************************************************************************
 */

char readbuf[255];      /* Buffer for reading from file                */
char lnbuf[255];        /* Input text line buffer                      */
char *txtPtr;           /* Pointer to next character to read in lnbuf  */

#define STACKSZ 8			/* Size of expression stacks   */
#define RETSTACKSZ 32			/* Size of return stack        */

int           operand_stack[STACKSZ];   /* Operand stack - grows down  */
unsigned char operator_stack[STACKSZ];  /* Operator stack - grows down */
int           return_stack[RETSTACKSZ]; /* Return stack - grows down   */

unsigned char operatorSP;               /* Operator stack pointer      */
unsigned char operandSP;                /* Operand stack pointer       */
unsigned char returnSP;                 /* Return stack pointer        */

jmp_buf       jumpbuf;                  /* For setjmp()/longjmp()      */

/*
 ***************************************************************************
 * Token table for expression parser
 ***************************************************************************
 */

/*
 * Single character binary operators - order must match binaryops string.
 * and must be sequential.
 */
const char binaryops[] = "^/%*+-><&#@";
#define TOK_POW     245 /* Binary ^              */
#define TOK_DIV     246 /* Binary /              */
#define TOK_MOD     247 /* Binary %              */
#define TOK_MUL     248 /* Binary *              */
#define TOK_ADD     249 /* Binary +              */
#define TOK_SUB     250 /* Binary -              */
#define TOK_GT      251 /* Binary >              */
#define TOK_LT      252 /* Binary <              */
#define TOK_BITAND  253 /* Binary &              */
#define TOK_BITOR   254 /* Binary # (| in C)     */
#define TOK_BITXOR  255 /* Binary @ (^ in C)     */

/*
 * Macro to determine if a token is a binary operator with one character.
 */
#define IS1CHBINARY(tok) ((tok >= TOK_POW) && (tok <= TOK_BITXOR))

/*
 * Two character binary operators - order must match binaryops1/2 strings.
 * and must be sequential.
 */
const char binaryops1[] = "=!><&#<>"; /* 1st char */
const char binaryops2[] = "====&#<>"; /* 2nd char */
#define TOK_EQL     237 /* Binary ==             */
#define TOK_NEQL    238 /* Binary !=             */
#define TOK_GTE     239 /* Binary >=             */
#define TOK_LTE     240 /* Binary <=             */
#define TOK_AND     241 /* Binary &&             */
#define TOK_OR      242 /* Binary ## (|| in C)   */
#define TOK_LSH     243 /* Binary <<             */
#define TOK_RSH     244 /* Binary >>             */

/*
 * Macro to determine if a token is a binary operator with two characters.
 */
#define IS2CHBINARY(tok) ((tok >= TOK_EQL) && (tok <= TOK_RSH))

/*
 * Unary operators - order must match unaryops string and must be sequential.
 * All unary operators are single character.
 */
const char unaryops[] = "-+!.*";
#define TOK_UNM     231 /* Unary -             */
#define TOK_UNP     232 /* Unary +             */
#define TOK_NOT     233 /* Unary !             */
#define TOK_BITNOT  234 /* Unary . (~ in C)    */
#define TOK_STAR    235 /* Unary * (deref.)    */

/*
 * Macro to determine if a token is a unary operator.
 */
#define ISUNARY(tok) ((tok >= TOK_UNM) && (tok <= TOK_STAR))

/*
 * Special token to mark end of stack
 */
#define SENTINEL    50

/*
 * Special token for illegal operator or statement.
 */
#define ILLEGAL     100

/*
 * Based on C's precedence rules.  Higher return value means higher precedence.
 * Ref: http://en.cppreference.com/w/c/language/operator_precedence
 */
unsigned char getprecedence(int token) {
	switch (token) {
		case TOK_UNP:
		case TOK_UNM:
		case TOK_STAR:
		case TOK_NOT:
		case TOK_BITNOT:
			return 11;
		case TOK_POW:
		case TOK_DIV:
		case TOK_MUL:
		case TOK_MOD:
			return 10;
		case TOK_ADD:
		case TOK_SUB:
			return 9;
		case TOK_LSH:
		case TOK_RSH:
			return 8;
		case TOK_GT:
		case TOK_GTE:
		case TOK_LT:
		case TOK_LTE:
			return 7;
		case TOK_EQL:
		case TOK_NEQL:
			return 6;
		case TOK_BITAND:
			return 5;
		case TOK_BITXOR:
			return 4;
		case TOK_BITOR:
			return 3;
		case TOK_AND:
			return 2;
		case TOK_OR:
			return 1;
		case SENTINEL:
			return 0; /* Must be lowest precedence! */
		default:
			/* Should never happen */
			EXIT(99);
	}
}

/*
 * Operator stack routines
 */

void push_operator_stack(unsigned char operator) {
	operator_stack[operatorSP] = operator;
	if (operatorSP == 0) {
		/* Warm start */
		print("?too cmplx");
		longjmp(jumpbuf, 1);
	}
	--operatorSP;
}

unsigned char pop_operator_stack() {
	if (operatorSP == STACKSZ - 1) {
		/* Warm start */
		longjmp(jumpbuf, 1);
	}
	++operatorSP;
	return operator_stack[operatorSP];
}

unsigned char top_operator_stack() {
	return operator_stack[operatorSP + 1];
}

/*
 * Operand stack routines
 */

void push_operand_stack(int operand) {
	operand_stack[operandSP] = operand;
	if (operandSP == 0) {
		/* Warm start */
		print("?too cmplx");
		longjmp(jumpbuf, 1);
	}
	--operandSP;
}

int pop_operand_stack() {
	if (operandSP == STACKSZ - 1) {
		/* Warm start */
		longjmp(jumpbuf, 1);
	}
	++operandSP;
	return operand_stack[operandSP];
}

int top_operand_stack() {
	return operand_stack[operandSP + 1];
}

/*
 ***************************************************************************
 * Parser proper ...
 ***************************************************************************
 */

/*
 * Binary operators
 *
 * Examines the character at *txtPtr, and if not NULL the character at
 * *(txtPtr+1).  If a two character binary operator is found, return the
 * token, otherwise if a one character binary operator is found return
 * the token.  If neither of these, return ILLEGAL.
 *
 * Uses the global tables binaryops (for single char operators) and
 * binaryops1 / binaryops2 (for two character operators).
 */
unsigned char binary() {
	unsigned char tok;
	unsigned char idx = 0;

	/*
	 * First two char ops (don't try if first char is NULL though!)
	 */

	if (*txtPtr != '\0') {
		tok = TOK_EQL;
		while (binaryops1[idx]) {
			if (binaryops1[idx] == *txtPtr) {
				if (binaryops2[idx] == *(txtPtr + 1)) {
					return tok;
				}
			}
			++idx;
			++tok;
		}
	}

	/*
	 * Now single char ops
	 */
	idx = 0;
	tok = TOK_POW;
	while (binaryops[idx]) {
		if (binaryops[idx] == *txtPtr) {
			return tok;
		}
		++idx;
		++tok;
	}

	return ILLEGAL;
}

/*
 * Unary operators
 *
 * Examines the character at *txtPtr.  If it is one of the unary operators
 * (which are all single character), returns the token value, otherwise
 * returns ILLEGAL.
 *
 * Uses the global tables unaryops (for single char operators).
 */
unsigned char unary() {
	unsigned char idx = 0;
	unsigned char tok = TOK_UNM;

	while (unaryops[idx]) {
		if (unaryops[idx] == *txtPtr) {
			return tok;
		}
		++idx;
		++tok;
	}
	return ILLEGAL;
}

/*
 * Pop an operator from the operator stack, pop the operands from the
 * operand stack and apply the operator to the operands.
 */
void pop_operator() {
	int operand2;
	int result;
	int token = pop_operator_stack();
	int operand1 = pop_operand_stack();

	if (!ISUNARY(token)) {

		/*
		 * Evaluate binary operator
		 * (Apply the operator token to operand1, operand2)
		 */
		operand2 = pop_operand_stack();

		switch (token) {
			case TOK_POW:
				result = _pow(operand2, operand1);
				break;
			case TOK_MUL:
				result = operand2 * operand1;
				break;
			case TOK_DIV:
				result = operand2 / operand1;
				break;
			case TOK_MOD:
				result = operand2 % operand1;
				break;
			case TOK_ADD:
				result = operand2 + operand1;
				break;
			case TOK_SUB:
				result = operand2 - operand1;
				break;
			case TOK_GT:
				result = operand2 > operand1;
				break;
			case TOK_GTE:
				result = operand2 >= operand1;
				break;
			case TOK_LT:
				result = operand2 < operand1;
				break;
			case TOK_LTE:
				result = operand2 <= operand1;
				break;
			case TOK_EQL:
				result = operand2 == operand1;
				break;
			case TOK_NEQL:
				result = operand2 != operand1;
				break;
			case TOK_AND:
				result = operand2 && operand1;
				break;
			case TOK_OR:
				result = operand2 || operand1;
				break;
			case TOK_BITAND:
				result = operand2 & operand1;
				break;
			case TOK_BITOR:
				result = operand2 | operand1;
				break;
			case TOK_BITXOR:
				result = operand2 ^ operand1;
				break;
			case TOK_LSH:
				result = operand2 << operand1;
				break;
			case TOK_RSH:
				result = operand2 >> operand1;
				break;
			default:
				/* Should never happen */
				EXIT(99);
		}

	} else {

		/*
		 * Evaluate unary operator
		 * (Apply the operator token to operand1)
		 */

		switch (token) {
			case TOK_UNM:
				result = -operand1;
				break;
			case TOK_UNP:
				result = operand1;
				break;
			case TOK_NOT:
				result = !operand1;
				break;
			case TOK_BITNOT:
				result = ~operand1;
				break;
			case TOK_STAR:
#ifndef __GNUC__
				result = PEEK(operand1);
#else
				print("*addr is a no-op on this platform.\n");
#endif
				break;
			default:
				/* Should never happen */
				EXIT(99);
		}
	}
	push_operand_stack(result);
}

void push_operator(int operator_token) {
	/* Handles operator precedence here */
	while (getprecedence(top_operator_stack()) >=
	         getprecedence(operator_token)) {

		pop_operator();
	}
	push_operator_stack(operator_token);
}

/*
 * Consume any space characters at txtPtr
 */
void eatspace() {
	while (*txtPtr == ' ') {
		++txtPtr;
	}
}

/* hint is set to two to signify a two character operator to be */
/* consumed.  Otherwise it is zero.                             */
/* TODO: THIS IS A MESS - GET RID OF IT!!                       */
void consume(int hint) {

	if (!*txtPtr) {
		/* End of buffer */
		return;
	}
	if (isalphach(*txtPtr)) {
		/* Consume variable name */
		while (isalphach(*txtPtr) || isdigitch(*txtPtr)) {
			++txtPtr;
		}
	} else if (isdigitch(*txtPtr)) {
		/* Consume entire value */
		/* TODO Don't think this is used anymore */
		while (isdigitch(*txtPtr)) {
			++txtPtr;
		}
	} else {
		/* Consume one byte operator */
		/* TODO This is ugly and should be fixed! */
		++txtPtr;
		if (hint == 2) {
			/* Consume second byte of two byte operator */
			++txtPtr;
		}
	}
	eatspace();
}

/*
 * Returns 0 on success, 1 if error
 */
unsigned char expect(unsigned char token) {

	if (*txtPtr == token) {
		consume(0);
		return 0;
	} else {
		if (token == 0) {
			print("?extra");
		} else {
			print("?expect ");
			printchar(token);
		}
		return 1;
	}
}

/*
 * Handles an expression
 * Returns 0 on success, 1 on error
 */
unsigned char E() {
	int op;

	if (P() != 0) {
		return 1;
	}

	while ((op = binary()) != ILLEGAL) {
		push_operator(op);
		if (IS1CHBINARY(op)) {
			consume(0);
		} else {
			consume(2);
		}
		if (P() != 0) {
			return 1;
		}
	}

	while (top_operator_stack() != SENTINEL) {
		pop_operator();
	}
	return 0;
}

/*
 * Handles a predicate
 * Returns 0 on success, 1 on error
 */
unsigned char P() {
	char key[5];
	int idx = 0;
	int arg = 0;

	if (*txtPtr == '\0') {
		return 1;
	}

	if (isalphach(*txtPtr)) {
		/*
		 * Handle variables
		 */
		while (isalphach(*txtPtr) || isdigitch(*txtPtr)) {
			if (arg < 4) {
				key[arg++] = *txtPtr;
			}
			++txtPtr;
		}
		key[arg] = '\0';
		
		if (*txtPtr == '[') {
			++txtPtr;
			if (eval(0, &idx) != 0) {
				return 1;
			}
			expect(']');
		}
		if (getintvar(key, idx, &arg) != 0) {
			return 1;
		}
		push_operand_stack(arg);
		//consume(0);
		eatspace();

	} else if (isdigitch(*txtPtr)) {
		/*
		 * Handle integer constants
		 */
		if (parseint(&arg) != 0) {
			return 1;
		}
		push_operand_stack(arg);
		//consume(0); /* parseint already advances txtPtr */
		eatspace();

	} else if (*txtPtr == '$') {
		/*
		 * Handle hex constants
		 */
		++txtPtr; /* Eat the $ */
		if (parsehexint(&arg) != 0) {
			return 1;
		}
		push_operand_stack(arg);
		//consume(0); /* parsehexint already advances txtPtr */
		eatspace();

	} else if (*txtPtr == '(') {
		/*
		 * Handle subexpressions in parenthesis
		 */
		consume(0);
		push_operator_stack(SENTINEL);
		if (E() != 0) {
			return 1;
		}
		if (expect(')') != 0) {
			return 1;
		}
		pop_operator_stack();

	} else if ((arg = unary()) != ILLEGAL) {
		/*
		 * Handle unary operator
		 */
		push_operator_stack(arg);
		consume(0);
		if (P() != 0) {
			return 1;
		}

	} else {
		/*
		 * Otherwise error
		 */
		print("?unexp");
		return 1;
	}
	return 0;
}

/*
 * Evaluate expression at txtPtr
 * If checkNoMore is 1 then expect(0) is called to check there is no extra
 * input to be consumed.
 * eval() is basically a wrapper around the expression parser routine E().
 * Result is returned via argument val.
 * Returns 0 if successful, 1 on error.
 */
unsigned char eval(unsigned char checkNoMore, int *val) {

	if (*txtPtr == 0) {
		print("?bad expr");
		return 1;
	}
	if (E() != 0) {
		return 1;
	}
	if (checkNoMore == 1) {
		if (*txtPtr == ';') {
			goto doret;
		}

		if (expect(0) != 0) {
			return 1;
		}
	}
doret:
	*val = pop_operand_stack();
	return 0;
}

/*
 * Everything above this line is the expression parser.
 * Everything below is the rest of the language implementation.
 */

char *heap1Ptr; /* BLK5 heap */
char *heap2Ptr; /* BLK3 heap */

#ifdef C64
/*
 * C64
 *
 * Here we have a continuous block of RAM from the top of the executable
 * to 0xbfff.  I retain the heap 1 / heap 2 convention of the VIC-20 for now.
 * For now I assign 8K to heap 1 and whatever is left for heap 2.
 *
 * Heap usage:
 *   Heap 1: Variables
 *   Heap 2: Program text
 */
#define HEAP1TOP (char*)0xbfff
#define HEAP1LIM (char*)0xa000

#define HEAP2TOP (char*)0x9fff - 0x0400 /* Leave $400 for the C stack */
#define HEAP2LIM (char*)0x4600 /* HEAP2LIM HAS TO BE ADJUSTED TO NOT
                                * TRASH THE CODE, WHICH LOADS FROM $0800 UP
                                * USE THE MAPFILE! */


#else

/*
 * VIC-20:
 *
 * We have two heaps because we have two discontinuous blocks of free memory
 * Heap 1: one using all of BLK5 (8KB)
 * Heap 2: growing down from the top of BLK3 (27.5K less size of executable)
 * The executable is around 12K at the time of writing, so there is approx
 * 12K free in heap2 once other overheads are taken into account.)
 *
 * Heap usage:
 *   Heap 1: Variables
 *   Heap 2: Program text
 */
#define HEAP1TOP (char*)0xbfff
#define HEAP1LIM (char*)0xa000

#define HEAP2TOP (char*)0x7fff - 0x0400 /* Leave $400 for the C stack */
#define HEAP2LIM (char*)0x5000 /* HEAP2LIM HAS TO BE ADJUSTED TO NOT
                                * TRASH THE CODE, WHICH LOADS FROM $1200 UP
                                * USE THE MAPFILE! */

#endif

/*
 * Clears heap 1.  Must call this before using alloc1().
 */
#define CLEARHEAP1() heap1Ptr = HEAP1TOP

/*
 * Clears heap 2.  Must call this before using alloc2().
 */
#define CLEARHEAP2() heap2Ptr = HEAP2TOP


/*
 * Allocate bytes on heap 1.
 */
void *alloc1(int bytes) {
#ifdef __GNUC__
	void *p = malloc(bytes);
	if (!p) {
		print("No mem!\n");
		longjmp(jumpbuf, 1);
	}
	return p;
#else
	if ((heap1Ptr - bytes) < HEAP1LIM) {
		print("No mem!\n");
		longjmp(jumpbuf, 1);
	}
	heap1Ptr -= bytes;
//printf("alloc1 %p\n", heap1Ptr);
	return heap1Ptr;
#endif
}

/*
 * Allocate bytes on heap 2.
 */
void *alloc2(int bytes) {
#ifdef __GNUC__
	void *p = malloc(bytes);
	if (!p) {
		print("No mem!\n");
		longjmp(jumpbuf, 1);
	}
	return p;
#else
	if ((heap2Ptr - bytes) < HEAP2LIM) {
		print("No mem!\n");
		longjmp(jumpbuf, 1);
	}
	heap2Ptr -= bytes;
//printf("alloc2 %p\n", heap2Ptr);
	return heap2Ptr;
#endif
}

/*
 * Return the total amount of free space on heap 1.
 */
int getfreespace1() {
	return (heap1Ptr - HEAP1LIM + 1);
}

/*
 * Return the total amount of free space on heap 2.
 */
int getfreespace2() {
	return (heap2Ptr - HEAP2LIM + 1);
}

/*
 * Set to 1 in immediate mode, 0 when running script
 */
unsigned char immediate = 1;

/*
 * Values:
 *  0 not editing program
 *  1 editing program
 *  2 editing program - insert first line
 * Only makes sense with immediate == 1
 */
unsigned char editmode = 0;

/*
 * Represents a line of VICScript code.
 */
struct lineofcode {
	char *line;
	struct lineofcode *next;
};

/*
 * Pointer to first line of code.
 */
struct lineofcode *program = NULL;

/*
 * Pointer to current line
 */
struct lineofcode *current = NULL;

/*
 * Used as a line number counter
 */
int counter;

/**
 * skipFlag is set to one when we enter a body of code which we are not
 * executing (for example because a while loop condition was false.)  When
 * skipFlag is one, the parser will only process certain loop control tokens
 * - all others are ignored.
 */
unsigned char skipFlag;

/*
 * Append a line to the program
 * If current is set, then the new line will be appended after current
 * and current will be moved forward to point to the newly added line.
 * If current is NULL, then this is the first line of a new program
 */
void appendline(char *line) {
	struct lineofcode *loc = alloc2(sizeof(struct lineofcode));

	loc->line = alloc2(sizeof(char) * (strlen(line) + 1));
	strcpy(loc->line, line);
	if (!current) {
		loc->next = NULL;
		new();
		program = current = loc;
	} else {
		loc->next = current->next;
		current->next = loc;
		current = loc;
	}
}

/*
 * Insert new first line (special case)
 */
void insertfirstline(char *line) {
	struct lineofcode *loc = alloc2(sizeof(struct lineofcode));

	loc->line = alloc2(sizeof(char) * (strlen(line) + 1));
	strcpy(loc->line, line);
	loc->next = program;
	program = loc;
}

/*
 * Make current point to the line with number linenum
 * (or NULL if not found).
 * Line numbers start from one in this routine.
 */
void findline(int linenum) {
	counter = 1;
	current = program;
	while (current) {
		if (counter == linenum) {
			return;
		}
		current = current->next;
		++counter;
	}
}

/*
 * Delete line(s)
 */
void deleteline(int startline, int endline) {
	int linesToDel = endline - startline + 1;
	struct lineofcode *prev = NULL;

	counter = 1;
	if (endline < startline) {
		return;
	}
	current = program;
	while (current && (linesToDel != 0)) {
		if (counter == startline) {
			if (prev) {
				prev->next = current->next;
			} else {
				program = current->next;
			}
#ifdef __GNUC__
			free(current->line);
			free(current);
#endif
			current = current->next; /* ILLEGAL BUT WORKS FOR NOW */
			--linesToDel;
			continue;
		}
		prev = current;
		current = current->next;
		++counter;
	}
}

/*
 * Replace line pointed to by current
 */
void changeline(char *line) {
#ifdef __GNUC__
	free(current->line);
#endif
	current->line = alloc2(sizeof(char) * (strlen(line) + 1));
	strcpy(current->line, line);
}

/*
 * Delete program, free memory.
 */
void new() {
#ifdef __GNUC__
	struct lineofcode *l = program;
	struct lineofcode *l2;

	while (l) {
		l2 = l->next;
		free(l->line);
		free(l);
		l = l2;
	}
#else
	/* No need to iterate and free them all, just dump the heap */
	CLEARHEAP2();
#endif
	program = NULL;
}

/*
 * Entry in the variable table
 * Uses first 4 characters as key (so variable and variety refer
 * to same item.)
 * For scalars the value is stored in value.  For arrays, the first
 * element is stored in value and the following values are stored
 * sequentially off the end of the struct.
 */
struct vartabent {
	char   name[5]; /* 4 chars plus NULL */
	struct vartabent *next;
	int    size;
	int    value;
};

struct vartabent *varsbegin;
struct vartabent *varsend;

/*
 * Make 4 byte key from variable name
 */
void getvarkey(char *name, char key[4]) {
	int i = strlen(name);
	key[0] = key[1] = key[2] = key[3] = 0;
	switch (i) {
		case 4:
			key[3] = name[3];
		case 3:
			key[2] = name[2];
		case 2:
			key[1] = name[1];
		case 1:
			key[0] = name[0];
			break;
		case 0:
			EXIT(99); // Bug if we get here
	}
}

/*
 * Find integer variable
 */
struct vartabent *findintvar(char *name) {
	char key[5];
	struct vartabent *ptr;

	getvarkey(name, key);
	key[4] = '\0';
	ptr = varsbegin;
	while (ptr) {
		if (strncmp(key, ptr->name, 4) == 0) {
			return ptr;
		}
		ptr = ptr->next;
	}
	return NULL; /* Not found */
}

/*
 * Clear all variables
 */
void clearvars() {
#ifdef __GNUC__
	struct vartabent *v2;
	struct vartabent *v = varsbegin;

	while (v) {
		v2 = v->next;
		free(v);
		v = v2;
	}
#else
	/* No need to iterate and free them all, just dump the heap */
	CLEARHEAP1();
#endif
	varsbegin = NULL;
	varsend = NULL;
}

/*
 * Print all variables as a table
 */
void printvars() {
	unsigned char i;
	struct vartabent *v = varsbegin;

	while (v) {
		print(v->name);
		if (v->size > 1) {
			printchar('[');
			printdec(v->size);
			printchar(']');
		}
		if (v->size == 1) {
			for (i = 0; i < (5 - strlen(v->name)); ++i) {
				printchar(' ');
			}
			printdec(v->value);
		}
		printchar('\n');
		v = v->next;
	}
}

/*
 * Create new integer variable
 * name is the variable name
 * sz is the size (1 for a scalar, >1 for an array)
 * value is the initializer
 * Return 0 on success, 1 if error.
 */
unsigned char createintvar(char *name, int sz, int value) {
	char key[5];
	int i;
	struct vartabent *ptr = findintvar(name);
	struct vartabent *v = NULL;

	if (ptr) {
		print("?var ");
		print(name);
		print(" defined");
		return 1;
	}
	if (sz < 1) {
		print("?dim");
		return 1;
	}
	v = alloc1(sizeof(struct vartabent) + (sz - 1) * sizeof(int));

	getvarkey(name, key);
	key[4] = '\0';
	strncpy(v->name, key, 4);
	v->value = value;
	v->size = sz;
	v->next = NULL;

	for (i = 1; i < sz; ++i) {
		*(&(v->value) + i) = value;
	}

	if (varsend) {
		varsend->next = v;
	}
	varsend = v;
	if (!varsbegin) {
		varsbegin = v;
	}
	return 0;
}

/*
 * Set existing integer variable
 * name is the variable name
 * idx is the index into an array (or 0 for scalar)
 * value is the value to set
 * Return 0 if successful, 1 on error
 */
unsigned char setintvar(char *name, int idx, int value) {
	struct vartabent *ptr = findintvar(name);
	int *p = &(ptr->value) + idx;

	if (!ptr) {
		print("?no var ");
		print(name);
		return 1;
	}
	if ((idx < 0) || (idx >= ptr->size)) {
		print("?subscript");
		return 1;
	}
	*p = value;
	return 0;
}

/*
 * Get existing integer variable
 * name is the variable name
 * idx is the index into an array (or 0 for scalars)
 * Returns the value in val.
 * Return 0 if successful, 1 on error
 */
unsigned char getintvar(char *name, int idx, int *val) {
	struct vartabent *ptr = findintvar(name);
	int *p = &(ptr->value) + idx;

	if (!ptr) {
		print("?no var ");
		print(name);
		return 1;
	}
	if ((idx < 0) || (idx >= ptr->size)) {
		print("?subscript");
		return 1;
	}
	*val = *p;
	return 0;
}

/*
 * Do integer or integer array declaration
 * Expects "name = expr" at txtPtr
 * Returns 0 if OK, 1 for error
 */
unsigned char doint() {
	int j;
	char name[5];
	int i = 0;

	eatspace();
	if (!txtPtr || !isalphach(*txtPtr)) {
		print("?var");
		return 1;
	}
	while (*txtPtr && (isalphach(*txtPtr) || isdigitch(*txtPtr))) {
		if (i < 4) {
			name[i++] = *txtPtr;
		}
		++txtPtr;
	}
	name[i] = '\0';

	i = 1; /* i will be the array size (1 for scalar) */
	if (*txtPtr == '[') {
		if (expect('[')) {
			return 1;
		}
		if (eval(0, &i) != 0) {
			return 1;
		}
		if (expect(']')) {
			return 1;
		}
	}

	eatspace();

	if (expect(leftarrow)) {
		return 1;
	}

	eatspace();

	if (eval(1, &j) != 0) {
		return 1;
	}
	return createintvar(name, i, j);
}

/*
 * Do integer assign to variable.
 * Handles parsing:
 *     "var <- expr"
 * or, "var[expr1] <- expr2"
 * Returns 0 if no error, 1 if error
 */
unsigned char dolet() {
	int j;
	char name[5];
	int i = 0;

	if (!txtPtr || !isalphach(*txtPtr)) {
		print("?var");
		return 1;
	}
	while (*txtPtr && (isalphach(*txtPtr) || isdigitch(*txtPtr))) {
		if (i < 4) {
			name[i++] = *txtPtr;
		}
		++txtPtr;
	}
	name[i] = '\0';

	i = 0; /* i will be the array size (1 for scalar) */
	if (*txtPtr == '[') {
		if (expect('[')) {
			return 1;
		}
		if (eval(0, &i) != 0) {
			return 1;
		}
		if (expect(']')) {
			return 1;
		}
	}

	eatspace();

	if (*txtPtr != leftarrow) {
		print("?syn");
		return 1;
	}

	++txtPtr;
	eatspace();

	if (eval(1, &j) != 0) {
		return 1;
	}
	if (setintvar(name, i, j) != 0) {
		return 1;
	}
	return 0;
}

#define CALLFRAME   0xfffe  /* Magic number for CALL stack frame    */
#define IFFRAME     0xfffd  /* Magic number for IF stack frame      */
#define FORFRAME    0xfffc  /* Magic number for FOR stack frame     */
#define WHILEFRAME  0xfffb  /* Magic number for WHILE stack frame   */

/*
 * Push line number (or other int) to return stack.
 */
void push_return(int linenum) {
//printf("push_return(%d) SP=%d before\n", linenum, returnSP);
	return_stack[returnSP] = linenum;
	if (returnSP == 0) {
		print("?too deep\n");
		longjmp(jumpbuf, 1);
	}
	--returnSP;
}

/*
 * Pop line number (or other int) from return stack.
 */
int pop_return() {
	if (returnSP == RETSTACKSZ - 1) {
		print("?empty\n");
		longjmp(jumpbuf, 1);
	}
	++returnSP;
//printf("pop_return() ->%d\n", return_stack[returnSP]);
	return return_stack[returnSP];
}

/*
 * Handle entry into FOR loop.
 * Handles parsing:
 *     "var <- expr : expr"
 * or, "var[expr1] <- expr2 : expr3"
 * Returns 0 if no error, 1 if error
 * TODO: One this is stable see how much memory can be saved by factoring
 * out code common to dolet() and dofor().
 */
unsigned char dofor() {
	int j;
	int k;
	char name[5];
	int i = 0;

	/*
	 * Search the stack to see if there is already a FOR stack frame
	 * with matching line number.  This is to handle the case where
	 * we jump out of a loop and then run into the FOR statement
	 * again.  We can't keep creating stack frames!
	 */
	int  p = returnSP;
	while (p <= RETSTACKSZ - 1) {
		if ((return_stack[p] == FORFRAME) &&
		    (return_stack[p - 1] == counter)) {
	 		/*
			 * Unwind the stack.
	 		 */
			returnSP = p;
			break;
		}
		++p;
	}

	while (*txtPtr && (isalphach(*txtPtr) || isdigitch(*txtPtr))) {
		if (i < 4) {
			name[i++] = *txtPtr;
		}
		++txtPtr;
	}
	name[i] = '\0';

	i = 0; /* i will be the array size (1 for scalar) */
	if (*txtPtr == '[') {
		if (expect('[')) {
			return 1;
		}
		if (eval(0, &i) != 0) {
			return 1;
		}
		if (expect(']')) {
			return 1;
		}
	}

	eatspace();

	if (*txtPtr != leftarrow) {
		print("?syn");
		return 1;
	}

	++txtPtr;
	eatspace();

	if (eval(0, &j) != 0) {
		return 1;
	}
	if (setintvar(name, i, j) != 0) {
		return 1;
	}

	eatspace();
	expect(':');
	eatspace();

	if (eval(1, &k) != 0) {
		return 1;
	}

	/*
	 * Place the following on the return stack:
	 *   Magic value FORFRAME to indicate FOR loop stack frame
	 *   Line counter for the for statement (here) (int)
	 *   Loop limit (int)
	 *   Pointer to loop control variable (int*)
	 */
	push_return(FORFRAME);
	push_return(counter);
	push_return(k);
	push_return((int)&(findintvar(name)->value) + i);

	return 0;
}

/*
 * Handle iterating or exiting the FOR loop.
 * Expects to find pointer to loop variable, loop limit
 * and line counter on the return stack.
 * Returns 0 if no error, 1 if error, 2 if immediate mode
 */
unsigned char doendfor() {

	if (return_stack[returnSP + 4] != FORFRAME) {
		print("?no for");
		return 1;
	}

	/*
	 * Compare loop control variable and limit
	 */
	if (*(int*)(return_stack[returnSP + 1]) <
	      return_stack[returnSP + 2]) {

		/*
		 * If loop not done, increment loop control var, jump back
		 * to line after FOR
		 */
		++(*(int*)(return_stack[returnSP + 1]));

		/*
		 * Special hack to handle immediate mode.  If the FOR was
		 * executed in immediate mode, jump back to start of lnbuf
		 * and then eat the first statement (assumed to be the FOR)
		 * This means that the FOR has to be the first thing in the
		 * line.
		 */
		if (return_stack[returnSP + 3] == -1) {
			txtPtr = lnbuf;
			while (*txtPtr && *txtPtr != ';') {
				++txtPtr;
			}
			++txtPtr;
			return 2; /* Immediate mode */
		}

		findline(return_stack[returnSP + 3] + 1);
		--counter; /* Findline uses 1-based linenums */
		if (!current) {
			/* Should never get here! */
			EXIT(99);
		}
		/*
		 * If there are multiple statements separated with semicolons
		 * then skip the first statement.  If semicolon was found then
		 * leave txtPtr pointing the statement following the semicolon
		 * and return 2 to have parseline() look at the rest of the
		 * line.  Otherwise return 0 so parseline advances to the
		 * following line.  This should enable single-line
		 * FOR ...; ...; ENDFOR loops to work.
		 */

		txtPtr = current->line;
		while (*txtPtr && *txtPtr != ';') {
			++txtPtr;
		}
		if (*txtPtr == ';') {
			++txtPtr;
			return 2;
		} else {
			return 0;
		}
		return 0;
	}

	/* Done looping, unwind stack */
	pop_return();
	pop_return();
	pop_return();
	pop_return();
	return 0;
}

void doif(unsigned char arg) {

	/*
	 * Search the stack to see if there is already an IF stack frame
	 * with matching line number.  This is to handle the case where
	 * we jump out of a loop and then run into the IF statement
	 * again.  We can't keep creating stack frames!
	 */
	int  p = returnSP + 1;
	while (p <= RETSTACKSZ - 1) {
		if ((return_stack[p] == IFFRAME) &&
		    (return_stack[p - 2] == counter)) {
	 		/*
			 * Unwind the stack.
	 		 */
			returnSP = p;
			break;
		}
		++p;
	}
	
	/*
	 * Place the following on the return stack:
	 *   Magic value IFFRAME to indicate IF loop stack frame
	 *   Status value as follows:
	 *    0: skipFlag was already set so not evaluating my argument
	 *    1: skipFlag was clear and I set it (condition false)
	 *    2: skipFlag was clear and I left it clear (condition true)
	 *   Line number
	 */
	push_return(IFFRAME);

	if (skipFlag != 0) {
		push_return(0);
	} else {
		if (arg == 0) {
			skipFlag = 1;
			push_return(1);
		} else {
			push_return(2);
		}
	}
	push_return(counter);
}

unsigned char doelse() {

	if (return_stack[returnSP + 3] != IFFRAME) {
		print("?no if");
		return 1;
	}

	/*
	 * If the matching IF statement had condition true then the
	 * value at return_stack[returnSP + 2] will be 2.  If the matching
	 * IF statement had condition false then the value will be 1.
	 */ 
	if (return_stack[returnSP + 2] == 2) {
		skipFlag = 1;
	} else if (return_stack[returnSP + 2] == 1) {
		skipFlag = 0;
	}
	return 0;
}

unsigned char doendif() {

	if (return_stack[returnSP + 3] != IFFRAME) {
		print("?no if");
		return 1;
	}

	/*
	 * If skipFlag was false when we hit the matching IF statement,
	 * then the value at return_stack[returnSP + 2] will be 1 or 2.
	 * In this case, clear skipFlag.
	 */
	if (return_stack[returnSP + 2] != 0) {
		skipFlag = 0;
	}

	pop_return();
	pop_return();
	pop_return();
	return 0;
}

void dowhile(unsigned char arg) {

	/*
	 * Search the stack to see if there is already an WHILE stack frame
	 * with matching line number.  This is to handle the case where
	 * we jump out of a loop and then run into the WHILE statement
	 * again.  We can't keep creating stack frames!
	 */
	int  p = returnSP + 1;
	while (p <= RETSTACKSZ - 1) {
		if ((return_stack[p] == WHILEFRAME) &&
		    (return_stack[p - 2] == counter)) {
	 		/*
			 * Unwind the stack.
	 		 */
			returnSP = p;
			break;
		}
		++p;
	}
	
	/*
	 * Place the following on the return stack:
	 *   Magic value WHILEFRAME to indicate WHILE loop stack frame
	 *   Status value as follows:
	 *    0: skipFlag was already set so not evaluating my argument
	 *    1: skipFlag was clear and I set it (condition false)
	 *    2: skipFlag was clear and I left it clear (condition true)
	 *   Line number
	 */
	push_return(WHILEFRAME);

	if (skipFlag != 0) {
		push_return(0);
	} else {
		if (arg == 0) {
			skipFlag = 1;
			push_return(1);
		} else {
			push_return(2);
		}
	}
	push_return(counter);
}

unsigned char doendwhile() {

	if (return_stack[returnSP + 3] != WHILEFRAME) {
		print("?no while");
		return 1;
	}

	/*
	 * If skipFlag was false when we hit the matching WHILE statement,
	 * then the value at return_stack[returnSP + 2] will be 1 (condition
	 * false) or 2 (condition true).  If the WHILE was false, then just
	 * set clear skipFlag, pop the stack and keep going.  If the WHILE
	 * was true, pop the stack and jump back to the WHILE test again.
	 */
	if (return_stack[returnSP + 2] == 1) {

		/* While was false */
		skipFlag = 0;

	} else if (return_stack[returnSP + 2] == 2) {

		/*
		 * Special hack to handle immediate mode.  If the FOR was
		 * executed in immediate mode, jump back to start of lnbuf
		 * This means that the WHILE has to be the first thing in the
		 * line.
		 */
		if (return_stack[returnSP + 1] == -1) {
			txtPtr = lnbuf;
			return 2; /* Immediate mode */
		}

		/* While was true */
		findline(return_stack[returnSP + 1]);
		--counter; /* Findline uses 1-based linenums */
		if (!current) {
			/* Should never get here! */
			EXIT(99);
		}
		txtPtr = current->line;
	}

	pop_return();
	pop_return();
	pop_return();

	return 0;
}

/*
 * Parse a decimal integer constant.
 * The text to parse is pointed to by txtPtr.
 * The result is placed in val.
 * If successful returns 0, otherwise 1.
 */
unsigned char parseint(int *val) {
	*val = 0;
	if (*txtPtr == 0) {
		return 1;
	}
	if (!isdigitch(*txtPtr)) {
		return 1;
	}
	do {
		*val *= 10;
		*val += *txtPtr - '0';
		++txtPtr;
	} while (isdigitch(*txtPtr));
	return 0;
}

/*
 * Return value of hex char
 */
unsigned char hexchar2val(char c) {
	if (c >= 'a' && c <= 'f') {
		return c - 'a' + 10;
	}
	return c - '0';
}

/*
 * Return character for hex digit 0 to 15
 */
#ifdef __GNUC__
char hexval2char(unsigned char val) {
	if (val > 9) {
		return val - 10 + 'a';
	}
	return val + '0';
}
#endif

/*
 * Parse a hexadecimal integer constant.
 * The text to parse is pointed to by txtPtr.
 * The result is placed in val.
 * If successful returns 0, otherwise 1.
 */
unsigned char parsehexint(int *val) {
	*val = 0;
	if (!(isdigitch(*txtPtr) || ((*txtPtr >= 'a') && (*txtPtr <= 'f')))) {
		return 1;
	}
	do {
		*val *= 16;
		*val += hexchar2val(*txtPtr);
		++txtPtr;
	} while (isdigitch(*txtPtr) || ((*txtPtr >= 'a') && (*txtPtr <= 'f')));
	return 0;
}

/*
 * Print a 16 bit integer value as an unsigned decimal
 */
void printdec(unsigned int val) {

#ifndef __GNUC__
	char buf[6];
	utoa(val, buf, 10);
	print(buf);
#else
	unsigned char digit;
	unsigned int denom = 10000;
	unsigned char doprint = 0;

	do {
		digit = val / denom;
		if (digit != 0) {
			doprint = 1;
		}
		if (doprint) {
			printchar(digit + '0');
		}
		val = val % denom;
		denom = denom / 10;
	} while (denom != 0);
	if (doprint == 0) {
		printchar('0');
	}
#endif
}

/*
 * Print a value as hex
 */
void printhex(unsigned int val) {

#ifndef __GNUC__
	char buf[6];
	printchar('$');
	utoa(val, buf, 16);
	print(buf);
#else
	printchar('$');
	printchar(hexval2char((val>>12) & 0x0f));
	printchar(hexval2char((val>>8) & 0x0f));
	printchar(hexval2char((val>>4) & 0x0f));
	printchar(hexval2char(val & 0x0f));
#endif
}

/*
 * Perform jump or call instruction
 * Expects label to jump to at txtPtr 
 * If param callmode is 1 then this is a subroutine call, otherwise it is
 * just a simple jump.
 * Return 0 if successful, 1 on error
 */
unsigned char dojump(unsigned char callmode) {
	char *p;
	char *oldTxtPtr;
	struct lineofcode *l = program;
	int origcounter = counter;

	counter = -1;
	while (l) {
		p = l->line;
		++counter;

		/*
		 * When JUMPing (but not CALLing), if the JUMP is ahead of the
		 * current line then parse those lines between the current
		 * line and the destination with skipFlag = 1.  This has the
		 * effect of ignoring the instructions (as if iterating
		 * through a WHILE or IF body where the condition is false)
		 * but processing statements that affect the stack ('if /
		 * else / endif / while / endw'.  The effect of this is to
		 * properly unwind the stack when jumping out of an inner
		 * loop.  For example:
		 *
		 * while i<10
		 *   while 1==1
		 *     if cond
		 *       jump done
		 *     endif
		 *   endw
		 *   lbl done
		 * endw
		 */   
		oldTxtPtr = txtPtr;
		if ((callmode != 1) && (counter > origcounter)) {
			skipFlag = 1;
			txtPtr = p;
			parseline(); 
		}
		skipFlag = 0;
		txtPtr = oldTxtPtr;

		while (p && (*p == ' ')) {
			++p;
		}
		if (strncmp(p, ((callmode == 1) ? "sub" : "lbl"), 3) == 0) {
			p += 3;
			while (p && (*p == ' ')) {
				++p;
			}
			if (strcmp(p, txtPtr) == 0) {

				if (callmode == 1) {
					/*
					 * For CALL, stack frame is just the
					 * magic number and the return line
					 */
					push_return(CALLFRAME);
					push_return(origcounter);
				}
				current = l;
				return 0;
			}
		}
		l = l->next;
	}
	print("?no ");
	print(((callmode == 1) ? "sub" : "lbl"));
	print(txtPtr);
	counter = origcounter;
	return 1;
}

/*
 * Handle return from subroutine.
 */
unsigned char doreturn() {
	/*
	 * Search the stack to find the first CALLFRAME.  This allows us
	 * to unwind any inner stackframes (for example where we return
	 * from within a FOR loop or IF statement.)
	 */
	int  p = returnSP + 1;
	while (p <= RETSTACKSZ - 1) {
		if (return_stack[p] == CALLFRAME) {
	 		/*
			 * Unwind the stack.
	 		 */
			returnSP = p;
			goto found;
		}
		++p;
	}

	print("?no call");
	return 1;

found:
	p = return_stack[p - 1];

	findline(p + 1);
	--counter; /* Findline uses 1-based linenums */
	if (!current) {
		/* Should never get here! */
		EXIT(99);
	}
	txtPtr = current->line;
	return 0;
}

/*
 * Statement tokens - must be in order with no gaps in sequence
 */
#define TOK_COMM    150 /* (comment) must be lowest-numbered token */
#define TOK_PRINTLN 151 /* println     */
#define TOK_PRHEXLN 152 /* prhexln     */
#define TOK_PRSTRLN 153 /* prstrln     */
#define TOK_PRINT   154 /* print       */
#define TOK_PRHEX   155 /* prhex       */
#define TOK_PRSTR   156 /* prstr       */
#define TOK_QUIT    157 /* quit        */
#define TOK_CLEAR   158 /* clear       */
#define TOK_VARS    159 /* vars        */
#define TOK_INT     160 /* int         */
#define TOK_RUN     161 /* run         */
#define TOK_NEW     162 /* new         */
#define TOK_LABEL   163 /* lbl         */
#define TOK_SUBR    164 /* sub         */
#define TOK_JUMP    165 /* jump        */
#define TOK_IF      166 /* if          */
#define TOK_ELSE    167 /* else        */
#define TOK_ENDIF   168 /* endif       */
#define TOK_FREE    169 /* free        */
#define TOK_POKE    170 /* poke (*)    */
#define TOK_CALL    171 /* call sub    */
#define TOK_RET     172 /* return      */
#define TOK_FOR     173 /* for         */
#define TOK_ENDFOR  174 /* endfor      */
#define TOK_WHILE   175 /* while       */
#define TOK_ENDW    176 /* endw        */
#define TOK_END     177 /* end         */

/* Line editor commands */
#define TOK_LOAD    178 /* Editor: load        */
#define TOK_SAVE    179 /* Editor: save        */
#define TOK_LIST    180 /* Editor: list        */
#define TOK_CHANGE  181 /* Editor: modify line */
#define TOK_APP     182 /* Editor: append line */
#define TOK_INS     183 /* Editor: insert line */
#define TOK_DEL     184 /* Editor: delete line */

/*
 * Used for the stmnttabent type field.  Code in parseline() uses this
 * value to determine how to handle parameters for the statement.
 *
 *  FULLLINE: full-line statement where the entire line 'belongs' to the
 *            statement (comments, sub and lbl are like this.)
 *  NOARGS: no arguments permitted.
 *  ONEARG: one expression is expected and evaluated.  No further arguments
 *          permitted.
 *  INITIALARG: one expression is evaluated.  Any subsequent arguments may be
 *              evaluated by custom code for each statement.
 *  TWOARGS: two expressions are expected, separated by a comma
 *  ONESTRARG: a string constant in quotes is expected
 *  CUSTOM: the statement has its own custom code to handle parameters
 */
enum stmnttype {FULLLINE,
                NOARGS,
                ONEARG,
                INITIALARG,
                ONESTRARG,
                CUSTOM};

/*
 * Represents an entry in the statement table
 */
struct stmnttabent {
	char *name;
	unsigned char token;
	enum stmnttype type;
};

/*
 * Statement table
 * Must be in order of sequentially increasing token value, so that
 * stmnttabent[t - TOK_COMM] is the line matching token t
 *
 * Also note that if a one name is a prefix of another, the longer one
 * must be first (so println goes before print).
 */
struct stmnttabent stmnttab[] = {

	/* Statements */
	{"\'",      TOK_COMM,           FULLLINE},
	{"println", TOK_PRINTLN,        ONEARG},
	{"prhexln", TOK_PRHEXLN,        ONEARG},
	{"prstrln", TOK_PRSTRLN,        ONESTRARG},
	{"print",   TOK_PRINT,          ONEARG},
	{"prhex",   TOK_PRHEX,          ONEARG},
	{"prstr",   TOK_PRSTR,          ONESTRARG},
	{"quit",    TOK_QUIT,           NOARGS},
	{"clear",   TOK_CLEAR,          NOARGS},
	{"vars",    TOK_VARS,           NOARGS},
	{"int",     TOK_INT,            CUSTOM},
	{"run",     TOK_RUN,            NOARGS},
	{"new",     TOK_NEW,            NOARGS},
	{"lbl",     TOK_LABEL,          FULLLINE},
	{"sub",     TOK_SUBR,           FULLLINE},
	{"jump",    TOK_JUMP,           FULLLINE},
	{"if",      TOK_IF,             ONEARG},
	{"else",    TOK_ELSE,           NOARGS},
	{"endif",   TOK_ENDIF,          NOARGS},
	{"free",    TOK_FREE,           NOARGS},
	{"*",       TOK_POKE,           INITIALARG},
	{"call",    TOK_CALL,           FULLLINE},
	{"return",  TOK_RET,            NOARGS},
	{"for",     TOK_FOR,            CUSTOM},
	{"endfor",  TOK_ENDFOR,         NOARGS},
	{"while",   TOK_WHILE,          ONEARG},
	{"endw",    TOK_ENDW,           NOARGS},
	{"end",     TOK_END,            NOARGS},

	/* Editor commands */
	{":r",      TOK_LOAD,           ONESTRARG},
	{":w",      TOK_SAVE,           ONESTRARG},
	{":l",      TOK_LIST,           CUSTOM},
	{":c",      TOK_CHANGE,         INITIALARG},
	{":a",      TOK_APP,            ONEARG},
	{":i",      TOK_INS,            ONEARG},
	{":d",      TOK_DEL,            INITIALARG}
};

/*
 * Attempt to find statement keyword
 * Returns the token or ILLEGAL
 *
 * Uses table stmnttab, returning the token corresponding to the first
 * matching name.
 */
unsigned char matchstatement() {
	unsigned char i;
	struct stmnttabent *s;

	for (i = 0; i < 35; ++i) {
		s = &(stmnttab[i]);
		if (strncmp(txtPtr, s->name, strlen(s->name)) == 0) {
			return s->token;
		}
	}
	return ILLEGAL;
}

/*
 * Used to check no arguments are passed to statements that do not take them
 * Returns 0 if end of line or semicolon next, 1 otherwise.
 */
unsigned char checkNoMoreArgs() {
	eatspace();
	if ((*txtPtr != '\0') && (*txtPtr != ';')) {
		print("?extra");
		return 1;
	}
	return 0;
}

/* Parse a line from the input buffer
 * Handles statements
 * Starts reading from location of txtPtr
 * Returns:
 *   0: Keep executing
 *   1: Normal stop
 *   2: Error stop
 */
unsigned char parseline() {
	int token;
	int arg;
	int arg2;
	char *p;
	struct stmnttabent *s;

	for(;;) {

startparse:
		eatspace();

		if (*txtPtr == 0) {
			break; /* Null statement */
		}
		token = matchstatement();

		/*
		 * If skipFlag is set, then only process those tokens that
		 * manipulate skipFlag:
		 *   'if / else / endif'
		 *   'while / endw'
		 * Skip all others.
		 */
		if (skipFlag != 0) {
			if ((token != TOK_IF) &&
			    (token != TOK_ELSE) &&
			    (token != TOK_ENDIF) &&
			    (token != TOK_WHILE) &&
			    (token != TOK_ENDW)) {

				/*
				 * Eat the statement up to semicolon or the
				 * end.
				 */
				while (*txtPtr && (*txtPtr != ';')) {
					++txtPtr;
				}
				goto nexttoken;
			}
		}

		if (token == ILLEGAL) {

			/*
			 * If the first character is a digit then treat
			 * this as an editor 'change line' command.  This
			 * allows the VIC-20 screen editor to work the same
			 * way as in BASIC.
			 */

			if (isdigitch(*txtPtr)) {

				token = TOK_CHANGE;
				s = &(stmnttab[token - TOK_COMM]);

			} else {

				/*
				 * Variable assignment winds up here
				 */
				if (dolet() != 0) {
					return 2; /* Error */
				}
				goto nexttoken;
			}
		} else {

			s = &(stmnttab[token - TOK_COMM]);

			/* Eat the keyword */
			txtPtr += strlen(s->name);

			eatspace();
		}

		/*
		 * Generic parameter handling based on statement type.
		 */
		switch (s->type) {

                	case NOARGS:
				/* Check end of input */
				arg = checkNoMoreArgs();
				if (arg != 0) {
					return 2;
				}
				break;
                	case ONEARG:
				/* Evaluate one arg and check end of input */
				if (eval(1, &arg) != 0) {
					return 2;
				}
				break;
			case INITIALARG:

				/* Evaluate one arg, don't check end of input */
				if (eval(0, &arg) != 0) {
					return 2;
				}
				break;
				
                	case ONESTRARG:
				/* Parse quoted string, place it in readbuf */
				if (!(*txtPtr == '"')) {
					return 2;
				}
				++txtPtr;
				p = readbuf;
				while (*txtPtr && (*txtPtr != '"')) {
					*(p++) = *(txtPtr++);
				}
				*p = '\0';
				if (*txtPtr == '"') {
					++txtPtr;
				} else {
					print("?unterm string");
					return 2;
				}
				break;
#ifdef __GNUC__
			case CUSTOM:
			case FULLLINE:
				break;
#endif
		}

		/*
		 * Code for individual statements.
		 */
		switch (token) {
			case TOK_COMM:
			case TOK_LABEL:
				/* Nothing to do */
				break;
			case TOK_SUBR:
				/* Error if we just run into this line! */
				print("?ran into sub");
				return 2;
				break;
			case TOK_QUIT:
#ifndef __GNUC__
#ifdef C64
				/* Restore normal NMI vector on C64 */
				POKE(808, 237);
#else
				/* Restore normal NMI vector on VIC20 */
				POKE(808, 112);
#endif
#endif

				print("Bye!\n");
				EXIT(0);
			case TOK_PRINT:
			case TOK_PRINTLN:
// TODO: Can do signed and unsigned here
				printdec(arg);
				if (token == TOK_PRINTLN) {
					printchar('\n');
				}
				break;
			case TOK_PRHEX:
			case TOK_PRHEXLN:
				printhex(arg);
				if (token == TOK_PRHEXLN) {
					printchar('\n');
				}
				break;
			case TOK_PRSTR:
			case TOK_PRSTRLN:
				print(readbuf);
				if (token == TOK_PRSTRLN) {
					printchar('\n');
				}
				break;
			case TOK_CLEAR:
				clearvars();
				break;
			case TOK_VARS:
				printvars();
				break;
			case TOK_INT:
				if (doint() != 0) {
					return 2; /* Error */
				}
				break;
			case TOK_RUN:
				run();
				break;
			case TOK_NEW:
				new();
				break;
			case TOK_JUMP:
			case TOK_CALL:
				if (dojump(token == TOK_CALL) != 0) {
					return 2;
				}
				break;
			case TOK_RET:
				if (doreturn() != 0) {
					return 2;
				}
				break;
			case TOK_FOR:
				if (dofor() != 0) {
					return 2;
				}
				break;
			case TOK_ENDFOR:
				/*
				 * Special hack to handle immediate mode.
				 */
				switch (doendfor()) {
					case 1: return 2;
					        break;
					case 2: goto startparse;
				}
				break;
			case TOK_WHILE:
				dowhile(!(arg == 0));
				break;
			case TOK_ENDW:
				/*
				 * Special hack to handle immediate mode.
				 */
				switch (doendwhile()) {
					case 1: return 2;
					        break;
					case 2: goto startparse;
				}
				break;
			case TOK_IF:
				doif(!(arg == 0));
				break;
			case TOK_ELSE:
				if (doelse() != 0) {
					return 2;
				}
				break;
			case TOK_ENDIF:
				if (doendif() != 0) {
					return 2;
				}
				break;
			case TOK_END:
				return 1;    /* Normal stop */
				break;
			case TOK_FREE:
#ifndef __GNUC__
				printdec(getfreespace1());
				print(" vars, ");
				printdec(getfreespace2());
				print(" code\n");
#else
				print("Plenty!\n");
#endif
				break;
			case TOK_POKE:
				eatspace();
				if (expect(leftarrow) != 0) {
					return 2;
				}
				eatspace();
				if (eval(1, &arg2) != 0) {
					return 2;
				}
#ifndef __GNUC__
				POKE(arg, arg2);
#else
				print("*addr_val not implemented on Linux.\n");
				print("You need a real computer (VIC-20)\n");
#endif
				break;
			case TOK_APP:
				findline(arg);
				if (!current) {
					print("?bad line\n");
					break;
				}
				editmode = 1;
				break;
			case TOK_INS:
				if (arg <= 1) {
					editmode = 2; /* Special mode for insert
       	                                                 first line */
				} else {
					findline(arg - 1);
					if (!current) {
						print("?bad line\n");
						break;
					}
					editmode = 1;
				}
				break;
			case TOK_DEL:
				eatspace();
				if (*txtPtr == 0) {
					deleteline(arg, arg); /* One arg */
					break;
				}
				if (expect(',') != 0) {
					return 2;
				}
				eatspace();
				if (eval(1, &arg2) != 0) {
					return 2;
				}
				deleteline(arg, arg2); /* Two args */
				break;
			case TOK_CHANGE:
				/* Bit of a hack to 'uneat' space eaten by
				 * eval() */
				while (*(txtPtr-1) == ' ') {
					--txtPtr;
				}
				findline(arg);
				if (!current) {
					print("?bad line\n");
					break;
				}
				changeline(txtPtr);
				break;
			case TOK_LIST:
				if (*txtPtr == 0) {
					list(1, 32767); /* No args */
					break;
				}
				if (eval(0, &arg) != 0) {
					return 2;
				}
				eatspace();
				if (*txtPtr == 0) {
					list(arg, 32767); /* One arg */
					break;
				}
				if (expect(',') != 0) {
					return 2;
				}
				eatspace();
				if (eval(1, &arg2) != 0) {
					return 2;
				}
				list(arg, arg2); /* Two args */
				break;
			case TOK_LOAD:
				if (readfile() != 0) {
					return 2; /* Error */
				}
				break;
			case TOK_SAVE:
				if (writefile() != 0) {
					return 2; /* Error */
				}
				break;
			default:
				/* Should never get here */
				EXIT(99);
		}

nexttoken:
		if (*txtPtr == ';') {
			++txtPtr;
			if (*txtPtr == '\0') {
				break;
			}
		} else {
			break;
		}
	}
	return 0;
}

/*
 * Expects filename in readbuf.
 * If writemode = 0 then it opens file for reading, otherwise for writing.
 * Returns 0 if OK, 1 if error.
 */
unsigned char openfile(unsigned char writemode) {
	char *readPtr = readbuf;

	if (writemode == 0) {
		print("Reading ");
	} else {
		print("Writing ");
	}
	print(readPtr);
	printchar(':');
	while(*readPtr != '\0') {
		++readPtr;
	}
	*(readPtr++) = ',';
	*(readPtr++) = 's';
	*readPtr = '\0';

	readPtr = readbuf;

/* File handling code is only for cc65 CBM for now */
/* TODO: Should also provide a POSIX version for GCC */
#ifndef __GNUC__

	if (cbm_open(1, 8, (writemode ? CBM_WRITE : CBM_READ), readPtr) != 0) {
		print("?file");
		return 1;
	}
#else
	return 0;
#endif
}

/*
 * Load program from file.
 * Expects filename in readbuf.
 * Returns 0 if OK, 1 if error.
 */
unsigned char readfile() {
#ifndef __GNUC__
	unsigned char i;
	unsigned char j;
	int bytes;
	unsigned char foundEOL;
	unsigned char bytesInBuf = 0;
	char *readPtr = readbuf;
	int donereading = 0;
#endif
	int count = 0;

	if (openfile(0) != 0) {
		return 1;
	}

	clearvars();
	new();

/* File handling code is only for cc65 CBM for now */
/* TODO: Should also provide a POSIX version for GCC */
#ifndef __GNUC__

	readPtr = readbuf;
	do {
		if (donereading == 0) {
#ifdef DEBUG_READFILE
			print("About to read ");
			printdec(255 - bytesInBuf);
			print(" bytes\n");
#endif
			bytes = cbm_read(1, readPtr, 255 - bytesInBuf);
			if (bytes == -1) {
				print("?read");
				cbm_close(1);
				return 1;
			}
			if (bytes == 0) {
				donereading = 1;
			}

			readPtr += bytes;
			bytesInBuf += bytes;

#ifdef DEBUG_READFILE
			print("Read ");
			printdec(bytes);
			print(" bytes\nBuf[");
			for (i = 0; i < bytesInBuf; ++i) {
				printchar(readbuf[i]);
			}
			print("]\n");
#endif
		}

		foundEOL = 0;
		for (i = 0; i < bytesInBuf; ++i) {
			if (readbuf[i] == 10 || readbuf[i] == 13) {
				strncpy(lnbuf, readbuf, i);
				lnbuf[i] = 0;
				for (j = i+1; j < bytesInBuf; ++j) {
					readbuf[j-i-1] = readbuf[j];
				}
				readPtr = readbuf + bytesInBuf - i - 1;
				bytesInBuf -= (i + 1);
				foundEOL = 1;
				break;
			}
		}

		if (foundEOL == 1) {
			txtPtr = lnbuf;
			if (count == 0) {
				insertfirstline(txtPtr);
				findline(1);
			} else {
				appendline(txtPtr);
			}
			++count;
		} else {
			if (bytesInBuf == 255) {
				print("?too long");
				cbm_close(1);
				return 1;
			}
			/* Handle last line with missing CRLF */
			if (donereading == 1 && bytesInBuf != 0) {
				readbuf[bytesInBuf] = '\0';
				appendline(readbuf);
				++count;
				break;
			}
		}
		
	} while (bytes != 0 || bytesInBuf != 0);

	cbm_close(1);
#endif

	printdec(count);
	print(" lines\n");
	return 0;
}

/*
 * Save program to file.
 * Expects filename in readbuf.
 * Returns 0 if OK, 1 if error.
 */
unsigned char writefile() {
#ifndef __GNUC__
	int bytes;
	int index;
#endif

	if (openfile(1) != 0) {
		return 1;
	}

/* File handling code is only for cc65 CBM for now */
/* TODO: Should also provide a POSIX version for GCC */
#ifndef __GNUC__

	current = program;
	while (current) {
		index = 0;
		for (index = 0; index < strlen(current->line); ++index) {
			bytes = cbm_write(1, current->line + index, 1);
			if (bytes == 0) {
				goto error;
			}
		}
		bytes += cbm_write(1, "\n", 1);
		if (bytes == 0) {
			goto error;
		}
		current = current->next;
	}

	cbm_close(1);
	print("OK\n");
	return 0;

error:
	cbm_close(1);
	print("?write");
	return 1;
#else
	return 0;
#endif
}

void run() {
	int status = 0;
	counter = 0;
	immediate = 0;
	clearvars();
	skipFlag = 0;
	returnSP = RETSTACKSZ - 1;
	current = program;
	while (current && (status == 0)) {
#ifndef __GNUC__
		/* Check for STOP key */
		if (PEEK(0x00c5) == 24) {
			status = 3;
			break;
		}
#endif
		txtPtr = current->line;
		status = parseline();
		current = current->next;
		++counter;
	}
	switch (status) {
		case 2:
			print(" err at ");
			printdec(counter);
			printchar('\n');
			returnSP = (RETSTACKSZ - 1);
			skipFlag = 0;
			break;
		case 3:
			print("Brk at ");
			printdec(counter);
			printchar('\n');
			returnSP = (RETSTACKSZ - 1);
			skipFlag = 0;
			break;
	}
	immediate = 1;
}

void list(unsigned int startline, unsigned int endline) {
	unsigned int count = 1;

	current = program;
	while (current) {
		if ((count >= startline) && (count <= endline)) {
#ifndef __GNUC__
			printchar(28);       /* Red */
			printchar(18);       /* Reverse On */
#endif
			printdec(count);
#ifndef __GNUC__
			printchar(144);      /* Black */
			printchar(146);      /* Reverse Off */
#endif
			print(current->line);
			printchar('\n');
		}
		++count;
		current = current->next;
	}
}

/*
 * Clear the operator and operand stacks prior to evaluating expression.
 */
void clearexprstacks() {
	operandSP = STACKSZ - 1;
	operatorSP = STACKSZ - 1;
	push_operator_stack(SENTINEL);
}

/*
 * Entry point.
 */
#ifdef __GNUC__
int
#else
void
#endif
main() {

#ifndef __GNUC__
//	/* Check for the existence of RAM at 0xa000 and add it to heap */
//	if (PEEK(0xa000) == POKE(0xa000, PEEK(0xa000)+1)) {
//		_heapadd ((void *) 0xa000, 0x2000);
//	}

#ifndef C64
	POKE(0x900f, 254); /* Nice color scheme */
#else
        char *border = (char*)0xd020;                                           
        char *background = (char*)0xd021;                                       
        *border = 6;
        *background = 7;
#endif

	printchar(147);      /* Clear */
	printchar(28);       /* Red */
	printchar(18);       /* Reverse On */

	/* Disable RUNSTOP/RESTORE */
	POKE(808, 100);
#endif

	returnSP = RETSTACKSZ - 1;
	varsbegin = NULL;
	varsend = NULL;
	program = NULL;
	current = NULL;

#ifndef C64
	/* Looks great in 22 cols! */
	print("*** VICScript v0.3 ****** (c)Bobbi, 2017 ***\n\n");
#else
	print("      ***    VICScript v0.3    ***      ");
	print("      ***    (c)Bobbi, 2017    ***      \n\n");
#endif

#ifndef __GNUC__
	printchar(144);      /* Black */
	printchar(146);      /* Reverse Off */
#endif
	print("Free Software.\n");
	print("Licenced under GPL.\n\n");

	CLEARHEAP1();
	CLEARHEAP2();

	/* Warm reset goes here */
	if (setjmp(jumpbuf) == 1) {
		print("Restart\n");
	}

	for(;;) {
		clearexprstacks();
		if (editmode != 0) {
#ifndef __GNUC__
			printchar(30);       /* Green */
			printchar(18);       /* Reverse On */
#endif
			printchar('>');
#ifndef __GNUC__
			printchar(144);      /* Black */
			printchar(146);      /* Reverse Off */
#endif
		}
		getln(lnbuf, 255);
		switch (editmode) {
			case 0: /* Not editing - immediate mode execute */
				txtPtr = lnbuf;
				current = NULL;
				counter = -1;
				switch (parseline()) {
					case 1: print("Stop\n");
						break;
					case 2: print(" err\n");
						break;
				}
				if (returnSP != (RETSTACKSZ -1)) {
					print("?Unbal\n");
					returnSP = (RETSTACKSZ - 1);
				}
				skipFlag = 0;
				break;
			case 1: /* Editing the program, period to escape */
				if(lnbuf[0] == '.') {
					editmode = 0;
				} else {
					appendline(lnbuf);
				}
				break;
			case 2: /* Special case for insert first line */
				insertfirstline(lnbuf);
				findline(1);
				editmode = 1;
				break;
		}
	}
}

